@@ -1,5 +1,5 @@
 /* Sylph-Searcher - full-text search program for Sylpheed
- * Copyright (C) 2007-2008 Sylpheed Development Team
+ * Copyright (C) 2007-2009 Sylpheed Development Team
  */
 
 #include <sylph/sylmain.h>
@@ -18,6 +18,11 @@
 
 #define PG_CONN(dbinfo) ((PGconn *)dbinfo->dbdata)
 
+#define PART_SEPARATOR		"------------------------------------------------------------------------"
+#define PART_SEPARATOR_LEN	72
+#define MAX_BODY_LEN		(128 * 7 * 1024)
+
+
 static DBInfo *conn;
 
 static gint verbose = 0;
@@ -29,6 +34,68 @@ static gint stat_skip_folderinfo = 0;
 static gint stat_error = 0;
 
 
+static gchar *make_valid_utf8(const gchar *str)
+{
+	const gchar *p = str, *next;
+	gchar *valid, *vp;
+
+	vp = valid = g_malloc(strlen(str) + 1);
+	while (*p != '\0') {
+		if (!g_utf8_validate(p, -1, &next)) {
+			if (next > p) {
+				memcpy(vp, p, next - p);
+				vp += (next - p);
+			}
+			*vp++ = '?';
+			p = g_utf8_find_next_char(next, NULL);
+		} else {
+			memcpy(vp, p, next - p);
+			vp += (next - p);
+			break;
+		}
+	}
+	*vp = '\0'; 
+
+	return valid;
+}
+
+static gint string_append_with_escape_line(GString *text, const gchar *str)
+{
+	const gchar *p = str, *next;
+	gchar *valid_str = NULL;
+	gint error = 0; 
+
+	if (!str)
+		return 0;
+
+	if (!g_utf8_validate(str, -1, NULL)) {
+		valid_str = make_valid_utf8(str);
+		p = valid_str;
+		error = -1;
+	}
+
+	if (!strncmp(p, PART_SEPARATOR, PART_SEPARATOR_LEN)) {
+		g_string_append(text, "- ");
+		g_string_append(text, PART_SEPARATOR);
+		p += PART_SEPARATOR_LEN;
+	}
+
+	while ((next = strstr(p, "\n" PART_SEPARATOR)) != NULL) {
+		next++;
+		g_string_append_len(text, p, next - p);
+		g_string_append(text, "- ");
+		g_string_append(text, PART_SEPARATOR);
+		p = next + PART_SEPARATOR_LEN;
+	}
+	if (*p != '\0')
+		g_string_append(text, p);
+
+	if (valid_str)
+		g_free(valid_str);
+
+	return error;
+}
+
 static gboolean is_nospace_ascii_str(const gchar *s, gint len)
 {
 	const gchar *p = s;
@@ -46,56 +113,205 @@ static gboolean is_nospace_ascii_str(const gchar *s, gint len)
 	return TRUE;
 }
 
-gchar *get_text_content(MsgInfo *msginfo)
-{
-	FILE *fp;
-	gchar *text;
-	gchar *p, *end;
-	gint possible_encline = 0;
-	gint prev_len = 0;
+#define MIN_B64LEN	64
+#define MIN_B64LINES	50
 
-	fp = procmime_get_first_text_content(msginfo, NULL);
-	if (!fp) {
-		g_warning("failed to get text content in msg %u",
-			  msginfo->msgnum);
-		return g_strdup("");
+static gboolean string_truncate_base64_lines(GString *text)
+{
+	gint len, prev_len = 0;
+	gchar *p, *end, *b64_start = NULL;
+	gint b64_lines = 0;
+	gboolean truncated = FALSE;
+
+	p = text->str;
+	while (*p != '\0') {
+		end = strchr(p, '\n');
+		if (end) {
+			len = end - p;
+			end++;
+		} else {
+			len = strlen(p);
+			end = p + len;
+		}
+
+		if (len >= MIN_B64LEN && is_nospace_ascii_str(p, len) &&
+		    (prev_len == 0 || prev_len == len)) {
+			if (!b64_start)
+				b64_start = p;
+			b64_lines++;
+			p = end;
+			prev_len = len;
+		} else {
+			if (b64_lines > MIN_B64LINES) {
+				gssize pos = b64_start - text->str;
+				gssize elen = p - b64_start;
+				gssize next = end - p;
+
+				g_string_erase(text, pos, elen);
+				p = text->str + pos + next;
+				truncated = TRUE;
+			} else {
+				p = end;
+			}
+
+			b64_start = NULL;
+			b64_lines = 0;
+			prev_len = 0;
+		}
 	}
 
-	text = file_read_stream_to_str(fp);
-	fclose(fp);
+	if (b64_lines > MIN_B64LINES) {
+		gssize pos = b64_start - text->str;
+		gssize elen = p - b64_start; 
 
-	if (!text) {
-		g_warning("failed to get text content from fp (msg %u)",
+		g_string_erase(text, pos, elen);
+		truncated = TRUE;
+	}
+
+	return truncated;
+}
+
+static gint string_truncate_lines(GString *text, gint maxlen)
+{
+	gint len = 0, llen;
+	gchar *p, *end;
+
+	if (text->len <= maxlen)
+		return 0;
+
+	string_truncate_base64_lines(text);
+	if (text->len <= maxlen)
+		return 1;
+
+	p = text->str;
+	while (*p != '\0') {
+		end = strchr(p, '\n');
+		if (!end) {
+			llen = strlen(p);
+			end = p + llen;
+		} else {
+			end++;
+			llen = end - p;
+		}
+		if (len + llen > maxlen) {
+			g_string_truncate(text, len);
+			return 1;
+		}
+
+		len += llen;
+		p = end;
+	}
+
+	return 1;
+}
+
+gchar *get_text_content(MsgInfo *msginfo, gint *attach_num)
+{
+	MimeInfo *mimeinfo;
+	MimeInfo *first_textpart;
+	MimeInfo *partinfo;
+	FILE *fp, *outfp;
+	GString *text;
+	gchar *str;
+
+	mimeinfo = procmime_scan_message(msginfo);
+	if (!mimeinfo) {
+		g_warning("procmime_scan_message failed: msg %u",
 			  msginfo->msgnum);
 		return NULL;
 	}
 
-	/* check for broken base64 contents */
-	p = text;
-	while (*p != '\0') {
-		end = strchr(p, '\n');
-		if (end) {
-			gint len = end - p;
-			if (len > 63 && is_nospace_ascii_str(p, len)) {
-				if (prev_len == len)
-					++possible_encline;
-				else
-					possible_encline = 1;
-				if (possible_encline > 80) {
-					g_warning("msg %u: possible base64-encoded lines detected! This message seems to be broken or PGP encrypted, so it will be ignored.", msginfo->msgnum);
-					g_free(text);
-					return NULL;
-				}
-			} else {
-				possible_encline = 0;
-			}
-			prev_len = len;
-			p = end + 1;
-		} else
-			break;
+	fp = procmsg_open_message(msginfo);
+	if (!fp) {
+		g_warning("procmsg_open_message failed: msg %u",
+			  msginfo->msgnum);
+		procmime_mimeinfo_free_all(mimeinfo);
+		return NULL;
 	}
 
-	return text;
+	first_textpart = mimeinfo;
+	while (first_textpart && first_textpart->mime_type != MIME_TEXT)
+		first_textpart = procmime_mimeinfo_next(first_textpart);
+	if (!first_textpart) {
+		first_textpart = mimeinfo;
+		while (first_textpart && first_textpart->mime_type != MIME_TEXT_HTML)
+			first_textpart = procmime_mimeinfo_next(first_textpart);
+	}
+
+	text = g_string_new("");
+
+	if (first_textpart) {
+		outfp = procmime_get_text_content(first_textpart, fp, NULL);
+		if (!outfp) {
+			g_warning("procmime_get_text_content failed: msg %u",
+				  msginfo->msgnum);
+			g_string_free(text, TRUE);
+			fclose(fp);
+			procmime_mimeinfo_free_all(mimeinfo);
+			return NULL;
+		}
+		str = file_read_stream_to_str(outfp);
+		if (str) {
+			string_append_with_escape_line(text, str);
+			g_free(str);
+		}
+		fclose(outfp);
+	}
+
+	*attach_num = 0;
+
+	for (partinfo = mimeinfo; partinfo != NULL;
+	     partinfo = procmime_mimeinfo_next(partinfo)) {
+		gchar *filename, *attname = NULL;
+
+		if (partinfo->mime_type == MIME_MULTIPART ||
+		    partinfo->mime_type == MIME_MESSAGE_RFC822)
+			continue;
+		if (first_textpart == partinfo)
+			continue;
+
+		filename = partinfo->filename ? partinfo->filename :
+			partinfo->name ? partinfo->name : NULL;
+		if (filename) {
+			*attach_num = *attach_num + 1;
+			attname = g_path_get_basename(filename);
+		}
+
+		if (partinfo->mime_type == MIME_TEXT ||
+		    partinfo->mime_type == MIME_TEXT_HTML) {
+			outfp = procmime_get_text_content(partinfo, fp, NULL);
+			if (!outfp) {
+				g_warning("procmime_get_text_content failed: msg %u", msginfo->msgnum);
+				g_free(attname);
+				fclose(fp);
+				procmime_mimeinfo_free_all(mimeinfo);
+				return g_string_free(text, FALSE);
+			}
+			str = file_read_stream_to_str(outfp);
+			if (str) {
+				g_string_append(text, "\n" PART_SEPARATOR "\n");
+				if (attname)
+					g_string_append_printf(text, "[ %s (%s %s) ]\n\n", attname, partinfo->content_type ? partinfo->content_type : "text", to_human_readable(partinfo->content_size));
+				string_append_with_escape_line(text, str);
+				g_free(str);
+			}
+			fclose(outfp);
+		} else {
+			if (attname) {
+				g_string_append(text, "\n" PART_SEPARATOR "\n");
+				g_string_append_printf(text, "[ %s (%s %s) ]\n\n", attname, partinfo->content_type ? partinfo->content_type : "application/octet-stream", to_human_readable(partinfo->content_size));
+			}
+		}
+
+		g_free(attname);
+	}
+
+	fclose(fp);
+	procmime_mimeinfo_free_all(mimeinfo);
+
+	string_truncate_lines(text, MAX_BODY_LEN);
+
+	return g_string_free(text, FALSE);
 }
 
 gint db_insert_msg(MsgInfo *msginfo)
@@ -136,12 +352,13 @@ gint db_insert_msg(MsgInfo *msginfo)
 		++stat_error;
 		return -1;
 	}
-	esc_folder_id = sql_escape_str(conn, folder_id);
 
 	if (db_start_transaction(conn) < 0) {
 		exit(1);
 	}
 
+	esc_folder_id = sql_escape_str(conn, folder_id);
+
 	msgid = sql_escape_str(conn, msginfo->msgid);
 	if (db_get_sid_from_msgid(conn, msgid, &sid) < 0) {
 		ret = -1;
@@ -159,6 +376,7 @@ gint db_insert_msg(MsgInfo *msginfo)
 		gchar *from, *to, *cc, *newsgroups, *subject, *inreplyto;
 		gchar *body_text, *esc_body_text;
 		gchar *body_index_text, *esc_body_index_text;
+		gint attach_num = 0;
 
 		if (msginfo->folder)
 			fullinfo = procmsg_msginfo_get_full_info(msginfo);
@@ -169,7 +387,7 @@ gint db_insert_msg(MsgInfo *msginfo)
 			++stat_error;
 			goto finish;
 		}
-		body_text = get_text_content(fullinfo);
+		body_text = get_text_content(fullinfo, &attach_num);
 		if (!body_text) {
 			if (msginfo->folder)
 				procmsg_msginfo_free(fullinfo);
@@ -179,6 +397,15 @@ gint db_insert_msg(MsgInfo *msginfo)
 		}
 		esc_body_text = sql_escape_str(conn, body_text);
 		body_index_text = get_wakachi_text(body_text);
+		if (!body_index_text) {
+			g_free(esc_body_text);
+			g_free(body_text);
+			if (msginfo->folder)
+				procmsg_msginfo_free(fullinfo);
+			ret = -1;
+			++stat_error;
+			goto finish;
+		}
 		esc_body_index_text = sql_escape_str(conn, body_index_text);
 
 		from = sql_escape_str(conn, fullinfo->from);
@@ -192,9 +419,9 @@ gint db_insert_msg(MsgInfo *msginfo)
 			folder_id, msginfo->msgnum, subject);
 
 		query = g_strdup_printf
-			("INSERT INTO msginfo(file_size, file_mtime, msg_date, flags, hdr_from, hdr_to, hdr_cc, hdr_newsgroups, hdr_subject, hdr_msgid, hdr_inreplyto, hdr_references, body_text, body_index)"
-		 	" VALUES(%u, E'%s', E'%s', %u, E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', to_tsvector(E'%s'))",
-		 	msginfo->size, mtime_str, date_str, msginfo->flags.perm_flags, from, to, cc, newsgroups, subject, msgid, inreplyto, "", esc_body_text, esc_body_index_text);
+			("INSERT INTO msginfo(file_size, file_mtime, msg_date, flags, hdr_from, hdr_to, hdr_cc, hdr_newsgroups, hdr_subject, hdr_msgid, hdr_inreplyto, hdr_references, body_text, body_index, attach_num)"
+		 	" VALUES(%u, E'%s', E'%s', %u, E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', E'%s', to_tsvector(E'%s'), %d)",
+		 	msginfo->size, mtime_str, date_str, msginfo->flags.perm_flags, from, to, cc, newsgroups, subject, msgid, inreplyto, "", esc_body_text, esc_body_index_text, attach_num);
 
 		res = PQexec(PG_CONN(conn), query);
 		g_free(query);
@@ -245,14 +472,14 @@ gint db_insert_msg(MsgInfo *msginfo)
 		if (sid == 0) {
 			query = g_strdup_printf("UPDATE msg_folderinfo SET msg_sid = currval('msginfo_msg_sid_seq') WHERE folder_id = E'%s' AND msgnum = %u", esc_folder_id, msginfo->msgnum);
 		} else {
-			query = g_strdup_printf("UPDATE msg_folderinfo SET msg_sid = %u WHERE folder_id = E'%s' AND msgnum = %u", sid, esc_folder_id, msginfo->msgnum);
+			query = g_strdup_printf("UPDATE msg_folderinfo SET msg_sid = %lu WHERE folder_id = E'%s' AND msgnum = %u", sid, esc_folder_id, msginfo->msgnum);
 		}
 	} else {
 		if (sid == 0) {
 			query = g_strdup_printf("INSERT INTO msg_folderinfo(msg_sid, folder_id, msgnum) VALUES(currval('msginfo_msg_sid_seq'), E'%s', %u)",
 						esc_folder_id, msginfo->msgnum);
 		} else {
-			query = g_strdup_printf("INSERT INTO msg_folderinfo(msg_sid, folder_id, msgnum) VALUES(%u, E'%s', %u)",
+			query = g_strdup_printf("INSERT INTO msg_folderinfo(msg_sid, folder_id, msgnum) VALUES(%lu, E'%s', %u)",
 						sid, esc_folder_id, msginfo->msgnum);
 		}
 	}
@@ -388,11 +615,9 @@ gint db_import_mh_folder(const gchar *path, gboolean del_nonexist,
 	GDir *dir;
 	Folder *mh_folder;
 	FolderItem *mh_item;
-	MsgInfo *msginfo;
 	GSList *mlist = NULL, *cur;
 	GSList *dirlist = NULL;
 	gchar *folder_id;
-	gint total = 0;
 	gint ret = 0;
 
 	g_return_val_if_fail(path != NULL, -1);
@@ -608,6 +833,20 @@ int main(int argc, char *argv[])
 			  options->user, options->pass);
 	if (!conn)
 		exit(1);
+	if (db_check_tables(conn) < 0) {
+		db_disconnect(conn);
+		exit(1);
+	}
+
+	if (options->delete_folder) {
+		if (db_delete_folder(conn, folder_id, TRUE) < 0) {
+			g_warning("remove folder failed: %s", folder_id);
+			db_disconnect(conn);
+			exit(1);
+		}
+		db_disconnect(conn);
+		exit(0);
+	}
 
 	db_import_folder(folder_id, del_nonexist, recursive);
 
