@@ -14,17 +14,21 @@
 #include <string.h>
 #include <time.h>
 #include <libpq-fe.h>
-#include <mecab.h>
+#ifdef HAVE_LIBMECAB
+#  include <mecab.h>
+#endif
 
 #include "common.h"
 
 #define PG_CONN(dbinfo) ((PGconn *)dbinfo->dbdata)
 
+#ifdef HAVE_LIBMECAB
 #ifdef G_OS_WIN32
 static CharSet mecab_encoding = C_SHIFT_JIS;
 #else
 static CharSet mecab_encoding = C_EUC_JP;
 #endif
+#endif /* HAVE_LIBMECAB */
 
 gchar *sql_escape_str(DBInfo *conn, const gchar *str)
 {
@@ -42,10 +46,36 @@ gchar *sql_escape_str(DBInfo *conn, const gchar *str)
 	return esc_str;
 }
 
+gchar *sql_escape_like_str(DBInfo *conn, const gchar *str)
+{
+	gchar *esc_like_str;
+	gchar *esc_str;
+	const gchar *p = str;
+	gchar *dp;
+	size_t len;
+
+	if (!str)
+		return g_strdup("");
+
+	len = strlen(str);
+	dp = esc_like_str = g_malloc(len * 2 + 1);
+	while (*p) {
+		if (*p == '%' || *p == '_' || *p == '\\')
+			*dp++ = '\\';
+		*dp++ = *p++;
+	}
+	*dp = '\0';
+	esc_str = sql_escape_str(conn, esc_like_str);
+	g_free(esc_like_str);
+
+	return esc_str;
+}
+
 #define MAXWAKACHILEN 128
 
 gchar *get_wakachi_text(const gchar *str)
 {
+#ifdef HAVE_LIBMECAB
 	gchar buf[MAXWAKACHILEN + 1];
 	mecab_t *mecab;
 	const mecab_node_t *node;
@@ -54,7 +84,18 @@ gchar *get_wakachi_text(const gchar *str)
 	GString *string;
 	gint len;
 
+#ifdef G_OS_WIN32
+	gchar *dir, *argv[2];
+
+	dir = g_win32_locale_filename_from_utf8(get_startup_dir());
+	argv[0] = "-r";
+	argv[1] = g_strdup_printf("%s\\mecabrc", dir);
+	mecab = mecab_new(2, argv);
+	g_free(argv[1]);
+	g_free(dir);
+#else
 	mecab = mecab_new2("");
+#endif
 	if (!mecab) {
 		g_warning("mecab_new2 returned error");
 		exit(1);
@@ -105,11 +146,56 @@ gchar *get_wakachi_text(const gchar *str)
 
 	g_string_free(string, TRUE);
 	return utf8_str;
+#else
+	return g_strdup(str);
+#endif /* HAVE_LIBMECAB */
+}
+
+GSList *get_phrase_text(const gchar *str)
+{
+	GSList *list = NULL;
+	const gchar *p = str;
+	const gchar *start;
+	const gchar *end;
+	gchar *phrase;
+
+	if (!str)
+		return NULL;
+
+	while ((p = strchr(p, '"'))) {
+		++p;
+		start = p;
+		if ((end = strchr(p, '"'))) {
+			if (start < end) {
+				phrase = g_strstrip(g_strndup(start, end - start));
+				g_print("phrase: %s\n", phrase);
+				if (*phrase)
+					list = g_slist_append(list, phrase);
+				else
+					g_free(phrase);
+			}
+			p = end + 1;
+		} else {
+			phrase = g_strstrip(g_strdup(start));
+			g_print("phrase: %s\n", phrase);
+			if (*phrase)
+				list = g_slist_append(list, phrase);
+			else
+				g_free(phrase);
+			break;
+		}
+	}
+
+	return list;
 }
 
 static void usage(const gchar *cmdname)
 {
+#ifdef HAVE_LIBMECAB
 	g_print("Usage: %s [-d dbname] [-h hostname] [-p port] [-U username] [-P password] [--mecab-encoding encoding] ...\n", cmdname);
+#else
+	g_print("Usage: %s [-d dbname] [-h hostname] [-p port] [-U username] [-P password] ...\n", cmdname);
+#endif
 	g_print("\n");
 
 	g_print("Common options:\n");
@@ -119,8 +205,10 @@ static void usage(const gchar *cmdname)
 	g_print("  -p port                       port number of database server\n");
 	g_print("  -U username                   username for database\n");
 	g_print("  -P password                   password for database\n");
+#ifdef HAVE_LIBMECAB
 	g_print("  --mecab-encoding encoding     encoding of MeCab dictionary\n");
 	g_print("                                (default: Unix: EUC-JP / Win32: Shift_JIS)\n");
+#endif
 	g_print("  --help                        show this message\n");
 	g_print("\n");
 
@@ -131,6 +219,7 @@ static void usage(const gchar *cmdname)
 	g_print("  -v                            verbose output\n");
 	g_print("  --exclude foldername          exclude foldername from import targets\n");
 	g_print("                                (can be specified multiple times)\n");
+	g_print("  --delete                      recursively delete folders from DB\n");
 #ifdef G_OS_WIN32
 	g_print("  --debug                       show debug console window\n");
 #endif
@@ -183,6 +272,7 @@ gint parse_cmdline(gint argc, gchar *argv[],
 				 *(p + 1) == 'v')
 				continue;
 			else if (*(p + 1) == '-') {
+#ifdef HAVE_LIBMECAB 
 				if (!strncmp(p + 2, "mecab-encoding", 14)) {
 					CharSet charset;
 
@@ -193,7 +283,9 @@ gint parse_cmdline(gint argc, gchar *argv[],
 						g_print("mecab_encoding = %s\n", argv[i + 1]);
 						mecab_encoding = charset;
 					}
-				} else if (!strncmp(p + 2, "help", 4))
+				} else
+#endif /* HAVE_LIBMECAB */
+				if (!strncmp(p + 2, "help", 4))
 					usage(argv[0]);
 			} else
 				usage(argv[0]);
@@ -243,6 +335,7 @@ gint parse_cmdline_options(gint argc, gchar *argv[], Options *options)
 			else if (*(p + 1) == 'v')
 				options->verbose = TRUE;
 			else if (*(p + 1) == '-') {
+#ifdef HAVE_LIBMECAB 
 				if (!strncmp(p + 2, "mecab-encoding", 14)) {
 					CharSet charset;
 
@@ -255,10 +348,14 @@ gint parse_cmdline_options(gint argc, gchar *argv[], Options *options)
 						g_print("mecab_encoding = %s\n", argv[i]);
 						mecab_encoding = charset;
 					}
-				} else if (!strncmp(p + 2, "exclude", 7)) {
+				} else
+#endif /* HAVE_LIBMECAB */
+				if (!strncmp(p + 2, "exclude", 7)) {
 					if (i + 1 == argc)
 						usage(argv[0]);
 					options->exclude_list = g_slist_append(options->exclude_list, g_strdup(argv[++i]));
+				} else if (!strncmp(p + 2, "delete", 6)) {
+					options->delete_folder = TRUE;
 				} else if (!strncmp(p + 2, "help", 4))
 					usage(argv[0]);
 			} else
@@ -369,6 +466,54 @@ gint db_disconnect(DBInfo *conn)
 	return 0;
 }
 
+gint db_check_connection(DBInfo *conn)
+{
+	if (!conn)
+		return -1;
+	if (PQstatus(PG_CONN(conn)) != CONNECTION_OK) {
+		g_warning("db_check_connection: connection bad");
+		return -1;
+	}
+
+	return 0;
+}
+
+gint db_check_tables(DBInfo *conn)
+{
+	PGresult *res;
+
+	if (!conn)
+		return -1;
+
+	res = PQexec(PG_CONN(conn), "SELECT table_name FROM information_schema.tables WHERE table_name = 'msginfo'");
+	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
+		db_error_message(conn, "SELECT FROM information_schema.tables failed");
+		PQclear(res);
+		return -1;
+	}
+	if (PQntuples(res) == 0) {
+		g_warning("table msginfo not found");
+		PQclear(res);
+		return -1;
+	}
+
+	res = PQexec(PG_CONN(conn), "SELECT column_name FROM information_schema. columns WHERE table_name = 'msginfo' AND column_name = 'attach_num'");
+	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
+		db_error_message(conn, "SELECT FROM information_schema.columns failed");
+		PQclear(res);
+		return -1;
+	}
+	if (PQntuples(res) > 0) {
+		PQclear(res);
+		return 0;
+	}
+	g_print("modifying msginfo table\n");
+	if (db_exec_command(conn, "ALTER TABLE msginfo ADD attach_num INTEGER") < 0)
+		return -1;
+
+	return 0;
+}
+
 void db_error_message(DBInfo *conn, const gchar *msg)
 {
 	gchar *pqmsg;
@@ -380,13 +525,13 @@ void db_error_message(DBInfo *conn, const gchar *msg)
 	g_free(pqmsg);
 }
 
-gint db_start_transaction(DBInfo *conn)
+gint db_exec_command(DBInfo *conn, const gchar *query)
 {
 	PGresult *res;
 
-	res = PQexec(PG_CONN(conn), "BEGIN");
+	res = PQexec(PG_CONN(conn), query);
 	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
-		db_error_message(conn, "BEGIN failed");
+		db_error_message(conn, "command failed");
 		PQclear(res);
 		return -1;
 	}
@@ -395,19 +540,19 @@ gint db_start_transaction(DBInfo *conn)
 	return 0;
 }
 
+gint db_start_transaction(DBInfo *conn)
+{
+	return db_exec_command(conn, "BEGIN");
+}
+
 gint db_end_transaction(DBInfo *conn)
 {
-	PGresult *res;
+	return db_exec_command(conn, "COMMIT");
+}
 
-	res = PQexec(PG_CONN(conn), "COMMIT");
-	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
-		db_error_message(conn, "COMMIT failed");
-		PQclear(res);
-		return -1;
-	}
-
-	PQclear(res);
-	return 0;
+gint db_abort_transaction(DBInfo *conn)
+{
+	return db_exec_command(conn, "ROLLBACK");
 }
 
 gint db_get_sid_from_msgid(DBInfo *conn, const gchar *msgid, gulong *sid)
@@ -471,7 +616,31 @@ gint db_delete_msg_from_folderinfo(DBInfo *conn, const gchar *folder_id,
 	gchar *query;
 	PGresult *res;
 
-	query = g_strdup_printf("DELETE FROM msg_folderinfo WHERE folder_id = E'%s' AND msgnum = %u", folder_id, msgnum);
+	if (msgnum == 0)
+		query = g_strdup_printf("DELETE FROM msg_folderinfo WHERE folder_id = E'%s'", folder_id);
+	else
+		query = g_strdup_printf("DELETE FROM msg_folderinfo WHERE folder_id = E'%s' AND msgnum = %u", folder_id, msgnum);
+	g_print("query: %s\n", query);
+	res = PQexec(PG_CONN(conn), query);
+	g_free(query);
+
+	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+		db_error_message(conn, "DELETE FROM msg_folderinfo failed");
+		PQclear(res);
+		return -1;
+	}
+
+	PQclear(res);
+	return 0;
+}
+
+gint db_delete_folderinfo_recursive(DBInfo *conn, const gchar *folder_id)
+{
+	gchar *query;
+	PGresult *res;
+
+	query = g_strdup_printf("DELETE FROM msg_folderinfo WHERE folder_id = E'%s' OR folder_id LIKE E'%s/%%'", folder_id, folder_id);
+	g_print("query: %s\n", query);
 	res = PQexec(PG_CONN(conn), query);
 	g_free(query);
 
@@ -490,7 +659,7 @@ gint db_is_sid_exist_in_folderinfo(DBInfo *conn, gulong sid)
 	gchar *query;
 	PGresult *res;
 
-	query = g_strdup_printf("SELECT msg_sid FROM msg_folderinfo WHERE msg_sid = %u LIMIT 1", sid);
+	query = g_strdup_printf("SELECT msg_sid FROM msg_folderinfo WHERE msg_sid = %lu LIMIT 1", sid);
 	res = PQexec(PG_CONN(conn), query);
 	g_free(query);
 
@@ -514,7 +683,7 @@ gint db_delete_msg(DBInfo *conn, gulong sid)
 	gchar *query;
 	PGresult *res;
 
-	query = g_strdup_printf("DELETE FROM msginfo WHERE msg_sid = %u", sid);
+	query = g_strdup_printf("DELETE FROM msginfo WHERE msg_sid = %lu", sid);
 	res = PQexec(PG_CONN(conn), query);
 	g_free(query);
 
@@ -527,3 +696,86 @@ gint db_delete_msg(DBInfo *conn, gulong sid)
 	PQclear(res);
 	return 0;
 }
+
+gint db_delete_msg_noref(DBInfo *conn)
+{
+	gchar *query;
+	PGresult *res;
+
+	query = g_strdup_printf("DELETE FROM msginfo WHERE msg_sid IN"
+				" (SELECT msg_sid FROM msginfo EXCEPT"
+				"  SELECT msg_sid FROM msg_folderinfo)");
+	g_print("query: %s\n", query);
+	res = PQexec(PG_CONN(conn), query);
+	g_free(query);
+
+	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+		db_error_message(conn, "DELETE FROM msginfo failed");
+		PQclear(res);
+		return -1;
+	}
+
+	PQclear(res);
+	return 0;
+}
+
+gint db_delete_folder(DBInfo *conn, const gchar *folder_id, gboolean recursive)
+{
+	gchar *esc_folder_id;
+	gint ret = 0;
+
+	if (db_start_transaction(conn) < 0)
+		return -1;
+
+	esc_folder_id = sql_escape_str(conn, folder_id);
+
+	if (recursive) {
+		if (db_delete_folderinfo_recursive(conn, esc_folder_id) < 0) {
+			ret = -1;
+			goto finish;
+		}
+	} else {
+		if (db_delete_msg_from_folderinfo(conn, esc_folder_id, 0) < 0) {
+			ret = -1;
+			goto finish;
+		}
+	}
+	if (db_delete_msg_noref(conn) < 0) {
+		ret = -1;
+		goto finish;
+	}
+
+finish:
+	if (db_end_transaction(conn) < 0)
+		ret = -1;
+
+	if (ret == 0)
+		g_print("removed %s\n", folder_id);
+
+	g_free(esc_folder_id);
+	return ret;
+}
+
+gchar *db_get_body(DBInfo *conn, gulong sid)
+{
+	PGresult *res;
+	gchar *query;
+	gchar *val, *text;
+
+	g_return_val_if_fail(conn != NULL, NULL);
+
+	query = g_strdup_printf("SELECT body_text FROM msginfo WHERE msg_sid = %lu", sid);
+	res = PQexec(PG_CONN(conn), query);
+	g_free(query);
+
+	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
+		db_error_message(conn, "SELECT failed");
+		return NULL;
+	}
+
+	val = PQgetvalue(res, 0, 0);
+	text = g_strdup(val);
+
+	PQclear(res);
+	return text;
+}
