@@ -49,6 +49,9 @@ gint searcher_config(AppConfig *config, GtkWindow *parent)
 		 GTK_STOCK_OK, GTK_RESPONSE_OK,
 		 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
 		 NULL);
+	gtk_window_set_position(GTK_WINDOW(dialog),
+				GTK_WIN_POS_CENTER_ON_PARENT);
+	gtk_window_set_transient_for(GTK_WINDOW(dialog), parent);
 	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
 	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
 	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
@@ -101,7 +104,7 @@ gint searcher_config(AppConfig *config, GtkWindow *parent)
 	}
 
 	gtk_widget_show_all(dialog);
-	gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(parent));
+	gtk_window_set_transient_for(GTK_WINDOW(dialog), parent);
 	result = gtk_dialog_run(GTK_DIALOG(dialog));
 
 #define UPDATE_MEMBER(member, entry) \
