@@ -32,6 +32,7 @@
 #include "version.h"
 #include "common.h"
 #include "searcher-config.h"
+#include "folder-selector.h"
 
 #define PG_CONN(dbinfo) ((PGconn *)dbinfo->dbdata)
 
@@ -53,6 +54,11 @@ typedef struct _QueryInfo {
 	gchar *subject;
 	gchar *date_gt;
 	gchar *date_lt;
+	gchar *folder;
+	gboolean folder_with_sub;
+	gboolean has_attach;
+
+	gint res_limit;
 } QueryInfo;
 
 static AppConfig config;
@@ -70,6 +76,10 @@ static struct SylFTSearchWindow {
 	GtkWidget *subject_entry;
 	GtkWidget *dategt_entry;
 	GtkWidget *datelt_entry;
+	GtkWidget *folder_entry;
+	GtkWidget *folder_btn;
+	GtkWidget *folder_with_sub_chkbtn;
+	GtkWidget *has_attach_chkbtn;
 
 	GtkWidget *treeview;
 	GtkListStore *store;
@@ -104,7 +114,6 @@ const gchar *ui_def =
 	"</ui>";
 
 static gint db_query(const QueryInfo *qinfo);
-static gchar *db_get_body_text(gulong sid);
 
 static void update_text(const gchar *text);
 
@@ -125,6 +134,9 @@ static void search_entry_activated	(GtkWidget	*widget,
 static gint search_deleted	(GtkWidget	*widget,
 				 GdkEventAny	*event,
 				 gpointer	 data);
+
+static void folder_btn_clicked	(GtkButton	*button,
+				 gpointer	 data);
 static void clear_clicked	(GtkButton	*button,
 				 gpointer	 data);
 static void search_clicked	(GtkButton	*button,
@@ -173,23 +185,43 @@ static gint db_query(const QueryInfo *qinfo)
 	query_info_free(query_info);
 	query_info = NULL;
 
+	if (db_check_connection(conn) < 0) {
+		gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label),
+				   _("Database connection error"));
+		return -1;
+	}
+
 	query = g_string_new("SELECT msg_sid, hdr_from, hdr_to, hdr_subject, msg_date, folder_id, msgnum FROM msginfo LEFT JOIN msg_folderinfo USING (msg_sid) WHERE ");
 
 	if (qinfo->body) {
 		gchar *sep_query_str;
+		GSList *phrase_text = NULL, *cur;
 
 		sep_query_str = get_wakachi_text(qinfo->body);
 		esc_query_str = sql_escape_str(conn, sep_query_str);
 		g_string_append_printf
-			(query, "query(E'%s') @@ body_index", esc_query_str);
+			(query, "plainto_tsquery(E'%s') @@ body_index",
+			 esc_query_str);
 		g_free(esc_query_str);
 		g_free(sep_query_str);
 		have_cond = TRUE;
+
+		phrase_text = get_phrase_text(qinfo->body);
+		if (phrase_text) {
+			for (cur = phrase_text; cur != NULL; cur = cur->next) {
+				esc_query_str = sql_escape_like_str
+					(conn, (gchar *)cur->data);
+				g_string_append_printf(query, " AND body_text ILIKE E'%%%s%%'", esc_query_str);
+				g_free(esc_query_str);
+				g_free(cur->data);
+			}
+			g_slist_free(phrase_text);
+		}
 	}
 
 #define MAKE_COND(header)						\
 	if (qinfo->header) {						\
-		esc_query_str = sql_escape_str(conn, qinfo->header);	\
+		esc_query_str = sql_escape_like_str(conn, qinfo->header); \
 		g_string_append_printf(query, "%shdr_" # header		\
 				       " ILIKE E'%%%s%%'",		\
 				       have_cond ? " AND " : "",	\
@@ -199,7 +231,17 @@ static gint db_query(const QueryInfo *qinfo)
 	}
 
 	MAKE_COND(from);
-	MAKE_COND(to);
+
+	if (qinfo->to) {
+		esc_query_str = sql_escape_like_str(conn, qinfo->to);
+		g_string_append_printf(query, "%s(hdr_to ILIKE E'%%%s%%'"
+				       " OR hdr_cc ILIKE E'%%%s%%')",
+				       have_cond ? " AND " : "",
+				       esc_query_str, esc_query_str);
+		g_free(esc_query_str);
+		have_cond = TRUE;
+	}
+
 	MAKE_COND(subject);
 
 	if (qinfo->date_gt) {
@@ -219,16 +261,34 @@ static gint db_query(const QueryInfo *qinfo)
 		have_cond = TRUE;
 	}
 
+	if (qinfo->folder) {
+		esc_query_str = sql_escape_str(conn, qinfo->folder);
+		if (qinfo->folder_with_sub) {
+			g_string_append_printf
+				(query, "%s(folder_id = E'%s' OR folder_id LIKE E'%s/%%')",
+				 have_cond ? " AND " : "", esc_query_str, esc_query_str);
+		} else {
+			g_string_append_printf
+				(query, "%sfolder_id = E'%s'",
+				 have_cond ? " AND " : "", esc_query_str);
+		}
+		g_free(esc_query_str);
+		have_cond = TRUE;
+	}
+	if (qinfo->has_attach && have_cond) {
+		g_string_append_printf(query, " AND attach_num > 0");
+	}
+
 	if (!have_cond) {
 		g_string_free(query, TRUE);
 		return -1;
 	}
 
-	if (config.res_limit > 0) {
+	if (qinfo->res_limit > 0) {
 		g_string_prepend(query, "SELECT * FROM (");
 		g_string_append_printf
 			(query, " ORDER BY msg_sid DESC LIMIT %d) AS m"
-				" ORDER BY msg_sid ASC", config.res_limit);
+				" ORDER BY msg_sid ASC", qinfo->res_limit);
 	}
 
 	g_print("query: %s\n", query->str);
@@ -237,6 +297,16 @@ static gint db_query(const QueryInfo *qinfo)
 
 	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
 		db_error_message(conn, "SELECT failed");
+		PQclear(res);
+		if (db_check_connection(conn) < 0) {
+			gtk_label_set_text
+				(GTK_LABEL(ftsearch_window.status_label),
+				 _("Database connection error"));
+		} else {
+			gtk_label_set_text
+				(GTK_LABEL(ftsearch_window.status_label),
+				 _("Database query error"));
+		}
 		return -1;
 	}
 
@@ -276,30 +346,6 @@ static gint db_query(const QueryInfo *qinfo)
 	return 0;
 }
 
-static gchar *db_get_body_text(gulong sid)
-{
-	PGresult *res;
-	gchar *query;
-	gchar *val, *text;
-
-	g_return_val_if_fail(conn != NULL, NULL);
-
-	query = g_strdup_printf("SELECT body_text FROM msginfo WHERE msg_sid = %lu", sid);
-	res = PQexec(PG_CONN(conn), query);
-	g_free(query);
-
-	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
-		db_error_message(conn, "SELECT failed");
-		return NULL;
-	}
-
-	val = PQgetvalue(res, 0, 0);
-	text = g_strdup(val);
-
-	PQclear(res);
-	return text;
-}
-
 static gboolean gtkut_text_buffer_match_string(GtkTextBuffer *textbuf,
 					       const GtkTextIter *iter,
 					       gunichar *wcs, gint len,
@@ -471,6 +517,10 @@ static void syl_searcher_create(void)
 	GtkWidget *dategt_entry;
 	GtkWidget *datelt_entry;
 	GtkTooltips *tip;
+	GtkWidget *folder_entry;
+	GtkWidget *folder_btn;
+	GtkWidget *folder_with_sub_chkbtn;
+	GtkWidget *has_attach_chkbtn;
 
 	GtkWidget *vpaned;
 
@@ -519,7 +569,7 @@ static void syl_searcher_create(void)
 	gtk_box_pack_start(GTK_BOX(vbox_top), vbox, TRUE, TRUE, 0);
 	gtk_container_set_border_width(GTK_CONTAINER (vbox), 8);
 
-	table = gtk_table_new(4, 2, FALSE);
+	table = gtk_table_new(5, 2, FALSE);
 	gtk_table_set_col_spacings(GTK_TABLE(table), 4);
 	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
 	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);
@@ -533,6 +583,11 @@ static void syl_searcher_create(void)
 	gtk_table_attach_defaults(GTK_TABLE(table), body_entry, 1, 2, 0, 1);
 	g_signal_connect(G_OBJECT(body_entry), "activate",
 			 G_CALLBACK(search_entry_activated), NULL);
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, body_entry,
+			     _("Enter words or sentence included in body text\n"
+			       "(Sentence is automatically divided into words)"),
+			     NULL);
 
 	hbox = gtk_hbox_new(FALSE, 8);
 	label = gtk_label_new(_("From:"));
@@ -543,6 +598,10 @@ static void syl_searcher_create(void)
 	gtk_table_attach_defaults(GTK_TABLE(table), from_entry, 1, 2, 1, 2);
 	g_signal_connect(G_OBJECT(from_entry), "activate",
 			 G_CALLBACK(search_entry_activated), NULL);
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, from_entry,
+			     _("Specify part of name or address of sender (From)"),
+			     NULL);
 
 	hbox = gtk_hbox_new(FALSE, 8);
 	label = gtk_label_new(_("To:"));
@@ -553,6 +612,10 @@ static void syl_searcher_create(void)
 	gtk_table_attach_defaults(GTK_TABLE(table), to_entry, 1, 2, 2, 3);
 	g_signal_connect(G_OBJECT(to_entry), "activate",
 			 G_CALLBACK(search_entry_activated), NULL);
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, to_entry,
+			     _("Specify part of name or address of recipient (To or Cc)"),
+			     NULL);
 
 	hbox = gtk_hbox_new(FALSE, 8);
 	label = gtk_label_new(_("Subject:"));
@@ -563,11 +626,37 @@ static void syl_searcher_create(void)
 	gtk_table_attach_defaults(GTK_TABLE(table), subject_entry, 1, 2, 3, 4);
 	g_signal_connect(G_OBJECT(subject_entry), "activate",
 			 G_CALLBACK(search_entry_activated), NULL);
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, subject_entry, _("Specify part of subject"),
+			     NULL);
+
+	hbox = gtk_hbox_new(FALSE, 8);
+	label = gtk_label_new(_("Folder:"));
+	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
+	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 4, 5, GTK_FILL, 0, 0, 0);
+
+	hbox = gtk_hbox_new(FALSE, 8);
+	gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 2, 4, 5);
+
+	folder_entry = gtk_entry_new();
+	gtk_box_pack_start(GTK_BOX(hbox), folder_entry, TRUE, TRUE, 0);
+	g_signal_connect(G_OBJECT(folder_entry), "activate",
+			 G_CALLBACK(search_entry_activated), NULL);
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, folder_entry, _("Specify search target folder"),
+			     NULL);
+
+	folder_btn = gtk_button_new_with_label("...");
+	gtk_box_pack_start(GTK_BOX(hbox), folder_btn, FALSE, FALSE, 0);
+
+	folder_with_sub_chkbtn = gtk_check_button_new_with_label(_("Search subfolders"));
+	gtk_box_pack_start(GTK_BOX(hbox), folder_with_sub_chkbtn, FALSE, FALSE, 0);
 
 	hbox = gtk_hbox_new(FALSE, 8);
 	label = gtk_label_new(_("Date newer than:"));
 	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
 	dategt_entry = gtk_entry_new();
+	gtk_widget_set_size_request(dategt_entry, 140, -1);
 	gtk_box_pack_start(GTK_BOX(hbox), dategt_entry, FALSE, FALSE, 0);
 	tip = gtk_tooltips_new();
 	gtk_tooltips_set_tip(tip, dategt_entry, _("Date format:\n"
@@ -578,13 +667,18 @@ static void syl_searcher_create(void)
 	label = gtk_label_new(_("older than:"));
 	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
 	datelt_entry = gtk_entry_new();
+	gtk_widget_set_size_request(datelt_entry, 140, -1);
 	gtk_box_pack_start(GTK_BOX(hbox), datelt_entry, FALSE, FALSE, 0);
 	tip = gtk_tooltips_new();
 	gtk_tooltips_set_tip(tip, datelt_entry, _("Date format:\n"
 						  "  2007/1/15\n"
 						  "  2007/02/01 12:34"),
 			     NULL);
-	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 4, 5, GTK_FILL, 0, 0, 0);
+
+	has_attach_chkbtn = gtk_check_button_new_with_label(_("Attachments"));
+	gtk_box_pack_start(GTK_BOX(hbox), has_attach_chkbtn, FALSE, FALSE, 0);
+
+	gtk_table_attach(GTK_TABLE(table), hbox, 0, 2, 5, 6, GTK_FILL, 0, 0, 0);
 
 	g_signal_connect(G_OBJECT(dategt_entry), "activate",
 			 G_CALLBACK(search_entry_activated), NULL);
@@ -683,6 +777,8 @@ static void syl_searcher_create(void)
 	GTK_WIDGET_SET_FLAGS(close_btn, GTK_CAN_DEFAULT);
 	gtk_box_pack_start(GTK_BOX(hbbox), close_btn, FALSE, FALSE, 0);
 
+	g_signal_connect(G_OBJECT(folder_btn), "clicked",
+			 G_CALLBACK(folder_btn_clicked), NULL);
 	g_signal_connect(G_OBJECT(search_btn), "clicked",
 			 G_CALLBACK(search_clicked), NULL);
 	g_signal_connect(G_OBJECT(clear_btn), "clicked",
@@ -701,6 +797,10 @@ static void syl_searcher_create(void)
 	ftsearch_window.subject_entry = subject_entry;
 	ftsearch_window.dategt_entry = dategt_entry;
 	ftsearch_window.datelt_entry = datelt_entry;
+	ftsearch_window.folder_entry = folder_entry;
+	ftsearch_window.folder_btn = folder_btn;
+	ftsearch_window.folder_with_sub_chkbtn = folder_with_sub_chkbtn;
+	ftsearch_window.has_attach_chkbtn = has_attach_chkbtn;
 	ftsearch_window.treeview = treeview;
 	ftsearch_window.store = store;
 	ftsearch_window.textview = textview;
@@ -844,6 +944,12 @@ int main(int argc, char *argv[])
 
 	syl_init();
 	locale_init();
+
+	prefs_common_read_config();
+	account_read_config_all();
+	if (folder_read_list() < 0)
+		g_warning("can't read folder list");
+
 	config_dir_init();
 
 	conf_ok = read_config(&config);
@@ -885,6 +991,9 @@ int main(int argc, char *argv[])
 	if (!conn)
 		alert_dialog(_("Connection to database failed"),
 			     _("Connection to database failed."));
+	else if (db_check_tables(conn) < 0)
+		alert_dialog(_("Checking of database failed"),
+			     _("Checking of database failed."));
 
 	gtk_main();
 
@@ -913,7 +1022,7 @@ static void selection_changed(GtkTreeSelection *selection, gpointer data)
 		return;
 	}
 
-	text = db_get_body_text(sid);
+	text = db_get_body(conn, sid);
 	update_text(text);
 	g_free(text);
 }
@@ -991,8 +1100,30 @@ static gint search_deleted(GtkWidget *widget, GdkEventAny *event,
 	return FALSE;
 }
 
+static void folder_btn_clicked(GtkButton *button, gpointer data)
+{
+	FolderItem *item;
+	gchar *id;
+
+	item = folder_selector(GTK_WINDOW(ftsearch_window.window));
+
+	if (!item || item->stype == F_VIRTUAL)
+		return;
+
+	id = folder_item_get_identifier(item);
+	if (id) {
+		gtk_entry_set_text(GTK_ENTRY(ftsearch_window.folder_entry), id);
+		g_free(id);
+	}
+}
+
 static void clear_clicked(GtkButton *button, gpointer data)
 {
+#if GTK_CHECK_VERSION(2, 6, 0)
+	gtk_tree_sortable_set_sort_column_id
+		(GTK_TREE_SORTABLE(ftsearch_window.store),
+		 GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, GTK_SORT_ASCENDING);
+#endif
 	gtk_list_store_clear(ftsearch_window.store);
 	update_text("");
 	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.body_entry), "");
@@ -1001,6 +1132,9 @@ static void clear_clicked(GtkButton *button, gpointer data)
 	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.subject_entry), "");
 	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.dategt_entry), "");
 	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.datelt_entry), "");
+	gtk_entry_set_text(GTK_ENTRY(ftsearch_window.folder_entry), "");
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ftsearch_window.folder_with_sub_chkbtn), FALSE);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ftsearch_window.has_attach_chkbtn), FALSE);
 	gtk_label_set_text(GTK_LABEL(ftsearch_window.status_label), "");
 	query_info_free(query_info);
 	query_info = NULL;
@@ -1029,6 +1163,16 @@ static void search_clicked(GtkButton *button, gpointer data)
 	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.datelt_entry));
 	if (str && *str)
 		qinfo.date_lt = (gchar *)str;
+	qinfo.res_limit = config.res_limit;
+	str = gtk_entry_get_text(GTK_ENTRY(ftsearch_window.folder_entry));
+	if (str && *str)
+		qinfo.folder = (gchar *)str;
+	if (gtk_toggle_button_get_active
+		(GTK_TOGGLE_BUTTON(ftsearch_window.folder_with_sub_chkbtn)))
+		qinfo.folder_with_sub = TRUE;
+	if (gtk_toggle_button_get_active
+		(GTK_TOGGLE_BUTTON(ftsearch_window.has_attach_chkbtn)))
+		qinfo.has_attach = TRUE;
 
 	db_query(&qinfo);
 }
