@@ -14,7 +14,8 @@ CREATE TABLE msginfo (
 	hdr_inreplyto TEXT,
 	hdr_references TEXT,
 	body_text TEXT,
-	body_index TSVECTOR
+	body_index TSVECTOR,
+	attach_num INTEGER
 );
 
 CREATE INDEX msginfo_msg_date_index ON msginfo (msg_date);
